unit game_unit;

{
  puzzle data, move registration, hints
}

interface

uses sysutils;

type TCellstatus = (clClosed,clOpen,clQueen);
     TCell = record
              status   : Tcellstatus;
              region   : byte;
              colorNr  : byte;
              Gcol     : byte;
              Grow     : byte;
              Qlev     : byte;  //# queens closing cell
              Hint     : byte; //0: nono; 1:X; 2:Queen
             end;
     TTestresult = record
                    Scode : byte;  //summary
                    Ecode : byte;  //error code
                    n1,n2 : byte;  //cells
                   end;
     THintResult = record
                    test   : byte;
                    nr     : byte;  //cell
                    column : byte;
                    row : byte;
                    region : byte;
                    crg   : word;  //column,row,region bit vector
                   end;
     TCRGstate = set of TcellStatus;
     TA4 = array[1..4] of byte;
     TA12 = array[1..12] of byte;

var game  : array[1..144] of TCell;
    gamesize : byte = 8;
    gamesize2 : byte = 64;
    hintresult : THintResult;
    HintActive : boolean;
    HintBusy   : boolean;
    HintCount : byte;

function makeRegions : Ttestresult;
function TestRegions : TTestresult;
function TestGame : TTestResult;
procedure NewGame;
procedure ChangeCell(n: byte);
procedure eraseCell(n : byte);
procedure OpenCell(n : byte);
procedure procHints;
procedure cell2CR(var C,R : byte; n : byte);
procedure addXX(n : byte);
procedure ApplyHints;
procedure CancelHints;

implementation

uses unit1;

procedure cell2CR(var C,R : byte; n : byte); //calc column,row of cell
begin
 C := ((n-1) mod gamesize) + 1;
 R := ((n-1) div gamesize) + 1;
end;

function CR2Cell(col,row : byte) : byte;
begin
 result := (row-1)*gamesize + col;
end;

procedure eraseCell(n : byte);
begin
 with game[n] do
  begin
   status := clOpen;
   Qlev := 0;
   hint := 0;
  end;
end;

function connected(a,b : byte) : boolean;  //Gcol,Grow gebruiken
var c1,c2,r1,r2 : byte;
begin
 result := false;
 if a=b then exit;
//
 c1 := (a-1) mod gamesize;
 r1 := (a-1) div gamesize;
 c2 := (b-1) mod gamesize;
 r2 := (b-1) div gamesize;
 result := (game[a].region = game[b].region) or
           (c1 = c2) or
           (r1 = r2) or
           ((abs(r1-r2) = 1) and (abs(c1-c2) = 1));
end;

procedure OpenCell(n: byte);
//open cell n
var i : byte;
begin
 with game[n] do
  if (status = clClosed) and (Qlev=0)  then
   begin
    status := clOpen;
    UpdateCell(n);
    exit;
   end;
 if game[n].status = clQueen then
  begin
   for i := 1 to gamesize2 do
    with game[i] do
     if connected(n,i) and (Qlev > 0) then
      begin
       dec(Qlev);
       if Qlev = 0 then begin
                         status := clOpen;
                         updateCell(i);
                        end;
      end;
   game[n].status := clOpen;
   UpdateCell(n);
  end;
end;

procedure ApplyHints;
var n : byte;
begin
 if HintCount < 250 then inc(HintCount);
 for n := 1 to gamesize2 do
  with game[n] do
   case hint of
    1 : begin
         status := clClosed;
         hint := 0;
         UpdateCell(n);
        end;
    2 : begin
         status := clQueen;
         hint := 0;
         UpdateCell(n);
        end ;
   end;//case
end;

procedure Cancelhints;
var n : byte;
begin
 for n := 1 to gamesize2 do
  with game[n] do
   if hint > 0 then
    begin
     hint := 0;
     updatecell(n);
    end;
end;

procedure addXX(n : byte);
//if queen added in cell n
var i : byte;
begin
 for i := 1 to gamesize2 do 
  if connected(n,i) then
   with game[i] do
    if status = clOpen then
     begin
      status := clClosed;
      Qlev := 1;
      updatecell(i);
     end else if Qlev > 0 then inc(Qlev);
end;

procedure ChangeCell(n: byte);
//called by mousedown on gamebox in play mode
begin
 with game[n] do
  case status of
   clOpen     : begin
                 status := clClosed;
                 UpdateCell(n);
                end;
   clClosed   : if Qlev = 0 then
                 begin
                  status :=clQueen;
                  if form1.Xcheckbox.Checked then AddXX(n);
                  UpdateCell(n);
                 end;
   clQueen    : OpenCell(n);
  end;//case
end;

procedure newGame;
//remove all data in cells
var n : byte;
begin
 for n := 1 to 144 do 
  with game[n] do
   begin
    status := clOpen;
    region := 0;
    Qlev := 0;
    hint := 0;
    colorNr := 0;
   end;
 form1.Caption := formheader;
end;

function MakeRegions : Ttestresult;
//exit Scode 0: OK; 1: unassigned cell; 2: colums <> regions
var i,k,n : byte;
    reg : array[1..15] of byte; //colorNr to region conversion
begin
 result.n1 := 0;
 result.n2 := 0;
 result.Scode := 0;
 k := 0;
 for i := 1 to 15 do reg[i] := 0;
 for n := 1 to gamesize2 do 
  begin
   i := Game[n].colorNr;
   if i = 0 then begin
                  result.n1 := n;
                  result.Scode := 1;
                  exit;             //unassigned cell
                 end;
   if reg[i] = 0 then begin         //new color detected
                       inc(k);      //next region
                       reg[i] := k;
                       Game[n].region := k;
                      end
    else Game[n].region := reg[i];  //old color
  end;
 if k <> gamesize then result.Scode := 2;
end;

function TestRegions : TTestresult;
//test disconnected regions
//result 0: OK; 1: disconnected regions;
var a,b,i,j,k,n,Xregion : byte;
    XGame : array[1..12,1..12] of byte;
    CF : boolean;//change flag
begin
 Xregion := 0;
 repeat   //all regions
  for j := 1 to 12 do for i := 1 to 12 do XGame[i,j] := 0;
  inc(XRegion);
  k := 0;
  for j := 1 to gamesize do       //init XGame
   for i := 1 to gamesize do
    begin
     n := CR2Cell(i,j);
     if Game[n].region = Xregion then begin
                                       inc(k);
                                       XGame[i,j] := k; //add sequence
                                      end
      else XGame[i,j] := 0; //not selected
    end;
  repeat                          //scan XGame
   CF := false;
   for j := 1 to gamesize do
    for i := 1 to gamesize  do
     begin
      a := XGame[i,j];
      if a > 0 then
       begin
        for k := 1 to 4 do
         begin b := 0;
          case k of
           1 : if i > 1 then b := XGame[i-1,j];
           2 : if i < gamesize then b := XGame[i+1,j];
           3 : if j > 1 then b := XGame[i,j-1];
           4 : if j < gamesize then b:= XGame[i,j+1];
          end;//case
          if (b > 0) and (a > b) then begin
                                      XGame[i,j] := b;
                                      CF := true;
                                     end;
         end;//for k
       end;// if a > 0
     end;//for i,j
  until CF = false;
  for j := 1 to gamesize do
   for i := 1 to gamesize do
    if XGame[i,j] > 1 then begin
                            result.n1 := CR2Cell(i,j);
                            result.Scode := 1;
                            exit;
                           end;
 until Xregion = gamesize;
 result.Scode := 0;
 result.n1 := 0;
end;

function testColumn(n : byte) : TCRGstate;
//return state of column n
//return empty set if multiple queens
var j,k,Qcount : byte;
begin
 result := [];
 Qcount := 0;
 for j := 1 to gamesize do
  begin
   k := CR2Cell(n,j);
   with game[k] do
    begin
     result := result + [status];
     if status = clQueen then inc(Qcount);
    end;
  end;
 if Qcount > 1 then result := [];  //if multiple queens
end;

function testRow(n : byte) : TCRGstate;
//return state of row n
//return empty set if multiple queens
var i,k,Qcount : byte;
begin
 result := [];
 Qcount := 0;
 for i := 1 to gamesize do
  begin
   k := CR2Cell(i,n);
   with game[k] do
    begin
     result := result + [status];
     if status = clQueen then inc(Qcount);
    end;
  end;
 if Qcount > 1 then result := [];
end;

function testRegion(n : byte) : TCRGstate;
//return state of region n
//return empty set if multiple queens
var k,Qcount : byte;
begin
 result := [];
 Qcount := 0;
 for k := 1 to gamesize2 do 
  with game[k] do
   if region = n then
    begin
     result := result + [status];
     if status = clQueen then inc(Qcount);
    end;
 if Qcount > 1 then result := [];
end;

procedure makeDiagonals(var a : TA4; n : byte);
//return cells left-top, right-top, left-bottom,right-bottom of cell n
var col,row,k,code : byte;
begin
 for k := 1 to 4 do a[k] := 0;
 code := 0;
 col := game[n].Gcol;
 row := game[n].Grow;
 if col = 1 then code := 1;
 if col = gamesize then code := code or 2;
 if row = 1 then code := code or 4;
 if row = gamesize then code := code or 8;
 if (code and 5) = 0 then a[1] := n - (gamesize+1);
 if (code and 6) = 0 then a[2] := n - (gamesize-1);
 if (code and 9) = 0 then a[3] := n + (gamesize-1);
 if (code and $a) = 0 then a[4] := n + (gamesize+1);
end;

function testGame : TTestResult;
//Scode    ---- ---1  OK
//              --1-  game is full
//              -1--  queens complete
//Ecode    ---- 0001  closed column n1
//              0010  closed row n1
//              0011  closed region n1
//              0101  multiple queens in column n1
//              0110                     row
//              0111                     region
//              1000  touching queens n1,n2
var k,n,m,ndiag : byte;
    crgstat : TCRGstate;
    diagonals : TA4;        //diagonal cells
    queencount,opencellcount : byte;
label L1;
begin
 with result do
  begin
   Scode := 0;
   Ecode := 0;
   n1 := 0;
   n2 := 0;
  end;
 queencount := 0;
 opencellcount := 0;
 for n := 1 to gamesize2 do
  with game[n] do
   case status of
    clOpen   : inc(opencellcount);
    clClosed : ;
    clQueen  : begin
                inc(queencount);
                makediagonals(diagonals,n);
                for k := 1 to 4 do
                 begin
                  m := diagonals[k];
                  if (m > 0) then
                   with result do
                    begin
                     ndiag := diagonals[k];
                     if (game[ndiag].status = clQueen) and (Ecode=0) then
                      begin
                       n1 := n;
                       n2 := ndiag;
                       Ecode := 8;
                      end;
                    end;
                 end;//for k
               end;//if status  clQueen
   end;//case
 with result do
  begin
   if opencellcount = 0 then Scode := 2; //set full game
   if queencount = gamesize then Scode := Scode or $4;
  end;
 n := 0;
 repeat
  inc(n); //count 1..gamesize
  k := 0;
  repeat
   inc(k);    //counts columns,rows,regions
   case k of
    1 : crgstat := testColumn(n);
    2 : crgstat := testRow(n);
    3 : crgstat := testRegion(n);
   end;//case
   if crgstat = [] then
    begin                         //multiple queens
     result.Ecode := result.Ecode or (k+4);
     result.n1 := n;
     goto L1;
    end;
   if crgstat = [clClosed] then   //closed
    begin
     result.Ecode := result.Ecode or k;
     result.n1 := n;
     goto L1;
    end;
  until k = 3;
 until n = gamesize;

L1:

  with result do
   if (Ecode = 0) then Scode := Scode or $1; //set OK bit
end;

//-- hint helpers --

function CombinationCount(n,k : byte) : word;
//return number of combinations if k choices from n
var d,i : byte;
begin
 if k > n-k then k := n-k;
 result := 1;
 d := 1;
 for i := n-k+1 to n do begin
                         result := (result * i) div d;
                         inc(d);
                        end;
end;

procedure Combination(var comdest:TA12; rank: word; n,k: byte);
//calculate combination "rank" of k choices from n elements.
//comdest array holds choosen elements (bytes: 1, 2,3....)
var el,ep,savedk : byte;
    dr : word;
begin
 el := 1;
 ep := 0;
 savedk := k; //save number of choices to make
 repeat
  dr := combinationCount(n-1,k-1);
  if rank < dr then begin
                     inc(ep);
                     dec(k);
                     comdest[ep] := el;
                    end
   else rank := rank - dr;
  dec(n);
  inc(el);
 until ep = savedk;
end;

function makeCRbits(var A : TA12; gsize : byte) : word;
//make bit vector of combination in A
var b,i : byte;
begin
 result := 0;
 for i := 1 to gsize do
  begin
   b := A[i];
   if b > 0 then result := result or (1 shl b);
  end;
end;

function checkColumnConnection(colnr,n : byte) : boolean;
//colnr: column number
//n: cell
//exit: true if all open cells of column connect to n
var i,n2 : byte;
    opencell : boolean;
begin
 result := true;
 n2 := colnr;
 i := 0;
 opencell := false;
 repeat
  inc(i);
  if game[n2].status = clOpen then
   begin
    opencell := true;
    result := connected(n,n2);
   end;
  n2 := n2 + gamesize;
 until (result = false) or (i = gamesize);
 result := result and opencell;
end;

function checkRowConnection(rowNr,n : byte) : boolean;
var i,n2 : byte;
    opencell : boolean;
begin
 n2 := (rowNr-1)*gamesize + 1;
 opencell := false;
 result := true;
 i := 0;
 repeat
  inc(i);
  if game[n2].status = clOpen then
   begin
    opencell := true;
    result := connected(n,n2);
   end;
  inc(n2);
 until (result = false) or (i =gamesize);
 result := result and opencell;
end;

function checkRegionConnection(regNr,n : byte) : boolean;
var n2,reg1,reg2 : byte;
    opencell : boolean;
begin
 result := true;
 n2 := 0;
 opencell := false;
 reg1 := game[n].region;
 repeat
  inc(n2);
  reg2 := game[n2].region;
  if (reg1 <> reg2) and (reg2 = regNr) and (game[n2].status = clOpen) then
   begin
    opencell := true;
    result := connected(n,n2);
   end;
 until (result = false) or (n2 = gamesize2); 
 result := result and opencell;
end;

procedure markOpenCells(n,k : byte);
//k=1: columns; k=2:rows; k=3: regions
//n: column, row, region
var cell,col,row,reg : byte;
    OK : boolean;
begin
 for cell := 1 to gamesize2 do 
  begin
   cell2CR(col,row,cell);
   reg := game[cell].region;
   case k of
    1 : OK := col=n;
    2 : OK := row=n;
    3 : OK := reg=n;
    else Ok := false;
   end;
   if OK and (game[cell].status = clOpen) then HmarkCell(cell,1);
  end;
end;

procedure MarkOpenDiagonals(n : byte);
var dd : TA4;
    k : byte;
begin
 makediagonals(dd,n);
 for k := 1 to 4 do
  begin
   n := dd[k];
   if (n > 0) and (game[n].status = clOpen) then HMarkCell(n,1);
  end;
end;

function GetSizeMask : word;
begin
 result := (1 shl (gamesize+1)) - 2;
end;

function Xscan(ColMask,RowMask,RegionMask : word) : boolean;
var i : byte;
    c,r,g : word;
begin
 result := false;
 for i := 1 to gamesize2 do 
  begin
   with game[i] do
    begin
     c := Gcol;
     r := Grow;
     g := region;
    end;
   if (((1 shl c) and ColMask) > 0) and
      (((1 shl r) and RowMask) > 0) and
      (((1 shl g) and RegionMask) > 0) then
    if game[i].status = clOpen then begin
                                     HMarkCell(i,1);
                                     result := true;
                                    end; 
  end;
end;

//-- hint procedures

procedure procHint1;  //open cells in col,row,reg with a queen  [11,12,13]
                      //multiple queens in col,row,reg          [1,2,3]
var n,k : byte;
    crgstate :TCRGstate;
begin
 n := 0;
 repeat               //check open cells in column,row,region with queen
  inc(n);
  k := 0;
  repeat
   inc(k);
   case k of
    1 : crgstate := testColumn(n);
    2 : crgstate := testRow(n);
    3 : crgstate := testRegion(n);
   end;//case
   if crgstate = [] then with hintresult do
                          begin
                           test := k;
                           nr := n;
                           case k of
                            1 : column := n;
                            2 : row := n;
                            3 : region := n;
                           end;
                          exit;               //error: multiple queens
                         end;
   if ([clQueen,clOpen] <= crgstate) then
    with hintresult do
     begin
      test := 10+k;
      case k of
       1 : column := n;
       2 : row := n;
       3 : region := n;
      end;
      MarkOpenCells(n,k);
      exit;
     end;
  until k = 3;
 until n = gamesize;
end;

procedure procHint2;
//diagonals of queen not closed
var c,k,n : byte;
    diag : TA4;
begin
 n := 0;
 repeat
  inc(n);
  if game[n].status = clQueen then
   begin
    makediagonals(diag,n);
    k := 0;
    repeat
     inc(k);
     c := diag[k];
     if (c > 0) and (game[c].status = clOpen) then
      with hintresult do
       begin
        test := 21;
        nr := n;
        markOpenDiagonals(n);
        exit;
       end;
    until k = 4;
   end;//if queen
 until n = gamesize2;
end;

procedure procHint31;
//X. Queen would close column
var cell,Xcolumn : byte;
begin
 for Xcolumn := 1 to gamesize do
  begin
   for cell := 1 to gamesize2 do
    begin
     if (game[cell].status=clOpen) and checkColumnConnection(Xcolumn,cell) then
      with hintresult do
       begin
        HMarkcell(cell,1);
        test := 31;
        nr := cell;
        column := Xcolumn;
       end;
    end;
   if hintresult.test > 0 then exit;
  end;
end;

procedure procHint32;
//x. Queen would close row
var cell,Xrow : byte;
begin
 for Xrow := 1 to gamesize do
  begin
   for cell := 1 to gamesize2 do 
    begin
     if (game[cell].status=clOpen) and checkRowConnection(Xrow,cell) then
      with hintresult do
       begin
        HMarkcell(cell,1);
        test := 32;
        nr := cell;
        row := Xrow;
       end;
    end;
   if hintresult.test > 0 then exit;
  end;
end;

procedure procHint33;
//x. Queen would close region
var cell,Xregion : byte;
begin
 for Xregion := 1 to gamesize do
  begin
   for cell := 1 to gamesize2 do 
    begin
     if (game[cell].status=clOpen) and checkRegionConnection(Xregion,cell) then
      with hintresult do
       begin
        HMarkcell(cell,1);
        test := 33;
        nr := cell;
        region := Xregion;
       end;
    end;
   if hintresult.test > 0 then exit;
  end;
end;

procedure procHint4;
//only open cell for queen in col,row,reg [41,42,43]
var Colc, Rowc, Regc : array[1..12] of byte; //open cell if 0<n<$80
    hit,k,n : byte;
    col,row,reg : byte;
begin
 for n := 1 to 12 do begin
                      Colc[n] := 0;
                      Rowc[n] := 0;
                      Regc[n] := 0;
                     end;
 for n := 1 to gamesize2 do 
  with game[n] do
   if status = clOpen then
    begin
     col := game[n].Gcol;
     row := game[n].Grow;
     Reg := region;
     if Colc[col] = 0 then Colc[col] := n else Colc[col] := $ff;
     if Rowc[row] = 0 then Rowc[row] := n else Rowc[row] := $ff;
     if Regc[reg] = 0 then Regc[reg] := n else RegC[reg] := $ff;
    end;
//
 n := 0;
 hit := 0;
 repeat       //n: 1..gamesize
  inc(n);
  for k := 1 to 3 do
   begin
    case k of
     1 : if (Colc[n] > 0) and (Colc[n] < $ff) then hit := Colc[n];
     2 : if (Rowc[n] > 0) and (Rowc[n] < $ff) then hit := Rowc[n];
     3 : if (Regc[n] > 0) and (Regc[n] < $ff) then hit := Regc[n];
    end;
    if hit > 0 then with hintresult do
                     begin
                      test := 40+k;
                      nr := hit;
                      case k of
                       1 : column := n;
                       2 : row := n;
                       3 : region := n;
                      end;
                      HmarkCell(hit,2);
                      exit;
                     end;
   end;
 until n = gamesize;
end;

procedure procHint51;
//processing columns containing regions
var regionColumns : array[1..12] of word;
    cell,col,i,reg,k : byte;
    maxColRank,ColRank,RegRank,selColumns,
    regcols,selRegions,FullMask,XregMask : word;
    A12,R12 : TA12;
begin
 fullmask := getSizeMask;
 for k := 1 to gamesize do regioncolumns[k] := 0;
 for cell := 1 to gamesize2 do 
  if game[cell].status = clOpen then
   begin
    col := game[cell].Gcol;
    reg := game[cell].region;
    regionColumns[reg] := regioncolumns[reg] or (1 shl col);
   end;
 k := 1;
 repeat          // selected # of columns = k
  inc(k);
  maxColRank := combinationcount(gamesize,k);
  ColRank := 0;
  repeat        // select k columns from gamesize
   combination(A12,ColRank,gamesize,k);
   SelColumns := makeCRBits(A12,k);
   RegRank := 0;
   repeat      // select k regions of gamesize
    combination(R12,RegRank,gamesize,k);
    XRegMask := makeCRbits(R12,k);
    selRegions := 0;
    for i := 1 to k do
     begin
      regcols := regionColumns[R12[i]];
      if regcols = 0 then selregions := $ffff
       else selregions := selregions or regCols;
     end;
    if selRegions = selColumns then
     if XScan(selColumns,fullMask,XRegMask xor fullMask) then
      with hintresult do
       begin
        test := 51;
        crg := selcolumns;
        exit;
       end;
    inc(RegRank);
   until RegRank = maxColrank;
   inc(ColRank);
  until ColRank = maxColRank;
 until k = gamesize-1;
end;

procedure procHint52;
//processing rows containing regions
var regionRows : array[1..12] of word;
    cell,row,i,reg,k : byte;
    maxRowRank,RowRank,RegRank,selRows,
    regRows,selRegions,FullMask,XregMask : word;
    A12,R12 : TA12;
begin
 fullmask := getSizeMask;
 for k := 1 to gamesize do regionRows[k] := 0;
 for cell := 1 to gamesize2 do 
  if game[cell].status = clOpen then
   begin
    row := game[cell].Grow;
    reg := game[cell].region;
    regionRows[reg] := regionRows[reg] or (1 shl row);
   end;
 k := 1;
 repeat          // selected # of columns = k
  inc(k);
  maxRowRank := combinationcount(gamesize,k);
  RowRank := 0;
  repeat        // select k columns from gamesize
   combination(A12,RowRank,gamesize,k);
   SelRows := makeCRBits(A12,k);
   RegRank := 0;
   repeat      // select k regions of gamesize
    combination(R12,RegRank,gamesize,k);
    selRegions := 0;
    for i := 1 to k do
     begin
      regRows := regionRows[R12[i]];
      if regRows = 0 then selRegions := $ffff
       else selregions := selregions or regRows;
     end;
    if selRegions = selRows then
     begin
      XRegMask := makeCRbits(R12,k) xor fullmask;
      if XScan(fullMask,selRows,XRegMask) then
       with hintresult do
        begin
         test := 52;
         crg := selRows;
         exit;
        end;
     end;
    inc(RegRank);
   until RegRank = maxRowrank;
   inc(RowRank);
  until RowRank = maxRowRank;
 until k = gamesize-1;
end;

procedure procHints;
//report in testresult
var i,k : byte;
label XF;
begin
 if HintBusy then exit;
//
 HintBusy := true;
 with hintresult do begin
                     test := 0;
                     nr   := 0;
                     column := 0;
                     row := 0;
                     region := 0;
                     crg  := 0;
                    end;
 k := 0;
 repeat
  inc(k);
  case k of
   1 : procHint1;
   2 : procHint2;
   3 : begin
        i :=0;
        repeat
         inc(i);
         case i of
          1 : procHint31;
          2 : procHint32;
          3 : procHint33;
         end;//case
         if hintresult.test > 0 then goto XF;
        until i=3;
       end;
   4 : procHint4;
   5 : begin
        i := 0;
        repeat
         inc(i);
         case i of
          1 : procHint51;
          2 : prochint52;
         end;
         if hintresult.test > 0 then goto XF;
        until i = 2;
       end;
  end;//case
 until (hintresult.test > 0) or (k = 5);

XF:

 hintBusy := false;
end;

end.
